#!/usr/bin/python

"""
A tree class for BayesForest.
This class imitates the capabilities of the tree-class of
the BayesForest Matlab interface.
See http://github.com/inuritdino/BayesForest
The main method to load a tree is by reading the MTG file
describing the tree's structure.
The class methods provide also for plotting of the structure
using matplotlib and calculating simple tree metrics like
girth, crown spread, and height.
Note: at the moment this class is used mainly for plotting
the tree structure and making publication quality figures.
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from scipy.linalg import norm
import re
#import time
# Font properties
font = {'family' : 'serif',
        'sans-serif' : 'Helvetica',
        'weight' : 'normal',
        'size' : 20}
mpl.rc('font', **font)

class tree:
    def __init__(self):
        self.capacity = 2000# not used at the moment
        self.n_alloc = 1# not used at the moment
        self.start = []#self.capacity * [None]
        self.end = []#self.capacity * [None]
        self.radius = []#self.capacity * [None]
        self.n_segments = 0
        return;

    def allocate(self):
        return;
        
    def add_segment(self,st,en,rad):
        self.start.append(st)
        self.end.append(en)
        self.radius.append(rad)
        self.n_segments += 1
        return;

    def plot_segment(self,ax,start,end,rad):
        #fig = plt.figure()
        #ax = fig.add_subplot(111, projection='3d')
        #origin = np.array([0, 0, 0])
        #axis and radius
        p0 = np.array(start) #np.array([1, 3, 2])
        p1 = np.array(end) #np.array([8, 5, 9])
        R = rad #5
        #vector in direction of axis
        v = p1 - p0
        #find magnitude of vector
        mag = norm(v)
        #unit vector in direction of axis
        v = v / mag
        #make some vector not in the same direction as v
        not_v = np.array([1, 0, 0])
        if (v == not_v).all():
            not_v = np.array([0, 1, 0])
            #make vector perpendicular to v
        n1 = np.cross(v, not_v)
        #normalize n1
        n1 /= norm(n1)
        #make unit vector perpendicular to v and n1
        n2 = np.cross(v, n1)
        #surface ranges over t from 0 to length of axis and 0 to 2*pi
        t = np.linspace(0, mag, 2)
        theta = np.linspace(0, 2 * np.pi, 10)
        #use meshgrid to make 2d arrays
        t, theta = np.meshgrid(t, theta)
        #generate coordinates for surface
        X, Y, Z = [p0[i] + v[i] * t + R * np.sin(theta) * n1[i] + R * np.cos(theta) * n2[i] for i in [0, 1, 2]]
        # Plot the cylinder surface in 3D
        ax.plot_surface(X, Y, Z, color = 'grey')

        return;

    def plot_tree(self,proj=None,save=None,title=None):
        fig = plt.figure(figsize=(10,9),dpi=100);
        ax = fig.add_subplot(111, projection='3d')
        # Setting the white background for the axis planes=panes
        ax.w_xaxis.set_pane_color((1.0,1.0,1.0,1.0))
        ax.w_yaxis.set_pane_color((1.0,1.0,1.0,1.0))
        ax.w_zaxis.set_pane_color((1.0,1.0,1.0,1.0))
        # Iteratively plot segments
        for i in xrange(self.n_segments):
            #print i
            self.plot_segment(ax,self.start[i],self.end[i],self.radius[i])
        ax.set_xlabel('X, m',labelpad=20)
        ax.set_ylabel('Y, m',labelpad=20)
        ax.set_zlabel('Z, m',labelpad=20)
        ax.set_zlim(zmin=0.0)
        if title:
            ax.set_title(title)
            
        if proj == None:
            ax.view_init(25,-55) # 3D view
        if proj == 'xz':
            ax.view_init(0,-90) #XZ
        if proj == 'xy':
            ax.view_init(90,-90) #XY
        if proj == 'yz':
            ax.view_init(0,0) #YZ
        if save == None:
            plt.show()
        else :
            plt.savefig(save,dpi=100)
        return;

    def read_mtg(self,filename,lpfg=True):
        dt = np.dtype([('segment',np.str_, 15),('radius', np.float64),('start',np.str_,60),
                       ('end',np.str_, 60)])
        C = np.loadtxt(filename,usecols=(0,4,5,6),dtype = dt)
        for i in xrange(len(C)):
            # Process the segment number
            #seg_idx = int(re.search('.*C(\d+)',C[i][0]).group(1))

            # Radius
            rad = C[i][1]

#            print C[i][3], C[i][2]
            # Start
            st = map(float,re.search('[\-0-9.eE]+,[\-0-9.eE]+,[\-0-9.eE]+',C[i][2]).group(0).split(',',2));
            
            if lpfg:
                # LPFG rotational system: X=x, Y=-z, Z=y
                st = [st[0], -st[2], st[1]]
            # End
            en = map(float,re.search('[\-0-9.eE]+,[\-0-9.eE]+,[\-0-9.eE]+',C[i][3]).group(0).split(',',2));
                
            if lpfg:
                # LPFG rotational system: X=x, Y=-z, Z=y
                en = [en[0], -en[2], en[1]]
            
            # Add the segment
            self.add_segment(st,en,rad)
            
        return C;

    def height(self):
        h = 0.0
        st = np.array(self.start)
        en = np.array(self.end)
        mxs = np.amax(st[:,2])
        mxe = np.amax(en[:,2])
        h = np.amax(np.array([mxs, mxe]))
        
        return h;

    def girth(self,girth_height=0.0):
        g = 2 * self.radius[0]

        return g;

    def crown_spread(self,da=10):
        c = 0.0
        deg = np.arange(-180.0,180.01,da)
        st = np.array(self.start)
        en = np.array(self.end)
        angles = (180/np.pi) * np.arctan2(en[:,1],en[:,0])
        center = np.array([st[0,0], st[0,1]])
        dists = np.sqrt( np.power( (en[:,0] - center[0]), 2) + np.power( (en[:,1] - center[1]), 2) )
        spokes = np.array([])
        for i in range(1,len(deg)):
            Jless = angles[:] < deg[i]
            Jmore = angles[:] > deg[i-1]
            J = np.logical_and(Jless,Jmore)
            if np.any(J == True):
                spokes = np.append(spokes, np.amax(dists[J]))

        c = 2 * np.mean(spokes)

        return c ;

# t = tree()
# t.read_mtg('Espoo.mtg',lpfg=False)
# t.plot_tree(proj='xy',save='Espoo.mtgXY.tif',title='Espoo QSM')

        

